import logging
import time
from logging import Logger
from queue import Queue

from .utilities import *
from .concepts import Event, EventHandler
from ..config import DataStore


class InternalPipelineElement(ABC):
    def __init__(self,
                 name: str = None,
                 data_store: DataStore = None,
                 logger: Logger = None,
                 logger_changers: List[Callable[[Logger], None]] = None,
                 input_flow_log_level=logging.DEBUG,
                 output_flow_log_level=logging.DEBUG,
                 ):
        self.name = get_not_none(name, self.__class__.__name__)
        self.data_store: DataStore = data_store
        self.logger: Logger = logger
        self.logger_changers: List[Callable[[Logger], None]] = get_not_none(logger_changers, [])
        self.input_flow_log_level = input_flow_log_level
        self.output_flow_log_level = output_flow_log_level

    def set_data_store(self, data_store: DataStore):
        self.data_store = data_store

    def set_logger(self, logger: Logger):
        for logger_changer in self.logger_changers:
            logger_changer(logger)
        self.logger = logger


class EventListener(InternalPipelineElement, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def process_event(self, event: Event) -> None:
        raise NotImplementedError()


class EventGenerator(InternalPipelineElement, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.event_handler: Optional[EventHandler] = None

    def process_generated_event(self, event: Event) -> None:
        if self.event_handler is None:
            raise Exception  # TODO
        else:
            self.event_handler.handle_event(event)


class EventSource(EventGenerator, ABC):
    pass


class TriggererEventSource(EventSource, ABC):
    @abstractmethod
    def start_generating(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def stop_generating(self) -> None:
        raise NotImplementedError()


class TriggerableEventSource(EventSource, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.priority: int = 1

    @abstractmethod
    def get_event(self) -> Optional[Event]:
        raise NotImplementedError()

    def get_event_generator_function(self) -> Optional[Callable[[], None]]:
        event: Optional[Event] = self.get_event()
        if event is None:
            return None
        return lambda: self.process_generated_event(event)


class TriggerableEventBatchSource(TriggerableEventSource, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.queue: Queue = Queue()

    @abstractmethod
    def get_event_batch(self) -> List[Event]:
        raise NotImplementedError()

    def get_event(self) -> Optional[Event]:
        if self.queue.empty():
            [self.queue.put(x) for x in self.get_event_batch()]
        if self.queue.empty():
            return None
        return self.queue.get(block=False)


class EventSink(EventListener):
    @abstractmethod
    def drown_event(self, event: Event) -> None:
        raise NotImplementedError()

    def process_event(self, event: Event) -> None:
        self.drown_event(event)


class EventAnalyzer(EventGenerator, EventListener):
    @abstractmethod
    def analyze_event(self, input_event: Event, output_event_handler: EventHandler) -> None:
        raise NotImplementedError()

    def process_event(self, event: Event) -> None:
        self.analyze_event(event, self.event_handler)



class KafkaServer:
    def __init__(self,
                 address: str,
                 port: int
                 ):
        self.address = address
        self.port = port

    def __str__(self):
        return f'{self.address}:{self.port}'


class KafkaTopic:
    def __init__(self,
                 server: KafkaServer,
                 topic_id: str
                 ):
        self.server = server
        self.topic_id = topic_id




import logging
from logging import Logger
from types import MethodType

import coloredlogs


def _modify_manager(manager):
    inform_parents = lambda node, child: (
        inform_parents(node.parent, (node if child is None else child)) if node.parent is not None else None,
        node.desc_added(node if child is None else child) if callable(getattr(node, 'desc_added', None)) else None,
        node)[2]
    if not hasattr(manager, "_modified"):
        manager._modified = True
        current_get_logger = manager.getLogger
        manager.getLogger = MethodType(lambda self, name: inform_parents(current_get_logger(name), None), manager)


class MaxHolder:
    def __init__(self, least=0):
        self._val = least

    def update_max(self, new_val):
        if new_val > self._val:
            self._val = new_val
        return self._val


def add_aligned_handler(logger: Logger, colored=True):
    _modify_manager(Logger.manager)
    max_holder = MaxHolder()
    formater = lambda child_name: \
        f'%(asctime)s [%(levelname)-8s] (%(name)-{max_holder.update_max(len(child_name))}s)   %(message)s'
    datefmt = "%Y-%m-%d %H:%M:%S"
    if colored:
        current_handle = [logging.StreamHandler()]
        logger.desc_added = lambda child: (
            logger.removeHandler(current_handle[0]),
            current_handle.remove(current_handle[0]),
            current_handle.append(get_colorful_handler(fmt=formater(child.name), datefmt=datefmt)),
            logger.addHandler(current_handle[0]))
    else:
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        logger.desc_added = \
            lambda child: handler.setFormatter(logging.Formatter(fmt=formater(child.name), datefmt=datefmt))
        logger.desc_added(logger)


def get_colorful_handler(fmt: str, datefmt: str) -> logging.Handler:
    logger = logging.getLogger("temp")
    coloredlogs.install(
        level=logging.NOTSET,
        logger=logger,
        field_styles=dict(
            asctime=dict(color='magenta'),
            hostname=dict(color='magenta'),
            levelname=dict(color='magenta', bold=True),
            name=dict(color='magenta'),
            programname=dict(color='magenta'),
            username=dict(color='magenta'),
        ),
        level_styles=dict(
            spam=dict(color='white', faint=True),
            debug=dict(color='white'),
            success=dict(color='white', bold=True),
            verbose=dict(color='white'),
            info=dict(color='cyan'),
            notice=dict(color='magenta'),
            warning=dict(color='yellow'),
            error=dict(color='red', faint=True),
            critical=dict(color='red', bold=True),
        ),
        fmt=fmt,
        datefmt=datefmt
    )
    return logger.handlers[0]


def get_default_logger(logger_name: str) -> Logger:
    result = logging.getLogger(logger_name)
    add_aligned_handler(result)
    result.setLevel(logging.DEBUG)
    logging.getLogger().handlers = result.handlers
    result.propagate = False
    return result



from .__private.concepts import Event, EventHandler
from .__private.elements import \
    InternalPipelineElement, \
    EventGenerator, EventListener, \
    EventSource, EventAnalyzer, EventSink, \
    TriggererEventSource, TriggerableEventSource, TriggerableEventBatchSource
from .__private.pipeline import Pipeline
